@@ -121,6 +121,8 @@ struct ncclTopoLinkList {
 #define RCCL_TOPO_FORCE_INTRA 16
 #define RCCL_TOPO_XGMI_ALL  32
 
+#define GCN_ARCH_NAME_LEN 16
+
 struct ncclTopoNode {
   int type;
   int64_t id;
@@ -131,7 +133,7 @@ struct ncclTopoNode {
       int rank;
       int cudaCompCap;
       int gdrSupport;
-      const char* gcn;
+      char gcn[GCN_ARCH_NAME_LEN];
       hipDeviceArch_t arch;
       int cu;
     }gpu;
@@ -190,6 +192,7 @@ struct ncclTopoSystem {
 
   // [RCCL] Track hostIdx to support rail-optimized rings/trees
   int hostIdx;
+  bool useRailOptimizedTrees;
 };
 
 ncclResult_t ncclTopoGetNode(struct ncclTopoSystem* system, struct ncclTopoNode** node, int type, uint64_t id);
@@ -250,7 +253,7 @@ static ncclResult_t ncclTopoIdToNetDev(struct ncclTopoSystem* system, int64_t id
       return ncclSuccess;
     }
   }
-  WARN("Could not find NET with id %lx\n", id);
+  WARN("Could not find NET with id %lx", id);
   return ncclInternalError;
 }
 
